@@ -1,6 +1,6 @@
 package ElasticSearch::Error;
 {
-  $ElasticSearch::Error::VERSION = '0.58';
+  $ElasticSearch::Error::VERSION = '0.59';
 }
 
 @ElasticSearch::Error::Internal::ISA       = __PACKAGE__;
@@ -23,7 +23,10 @@ package ElasticSearch::Error;
 use strict;
 use warnings FATAL => 'all', NONFATAL => 'redefine';
 
-use overload ( '""' => 'stringify' );
+use overload (
+    '""'  => 'stringify',
+    'cmp' => 'compare',
+);
 use Data::Dumper;
 
 #===================================
@@ -48,6 +51,15 @@ sub stringify {
     return $msg;
 }
 
+#===================================
+sub compare {
+#===================================
+    my ( $error, $other, $swap ) = @_;
+    $error .= '';
+    ( $error, $other ) = ( $other, $error ) if $swap;
+    return $error cmp $other;
+}
+
 =head1 NAME
 
 ElasticSearch::Error - Exception objects for ElasticSearch
